SYD(7)

# NAME

Overview of sandboxing with Syd

# SANDBOXING

There are fourteen sandboxing types:

. *Read sandboxing*
. *Write sandboxing* (and *Path Masking*)
. *Stat sandboxing* (aka *Path Hiding*)
. *Exec sandboxing* (and *SegvGuard*)
. *Ioctl sandboxing* (access *PTY*, *DRM*, *KVM* etc. safely)
. *Force sandboxing* (aka *Verified Execution*, like *Veriexec* & *IntegriForce*)
. *TPE sandboxing* (aka *Trusted Path Execution*)
. *Network sandboxing* (feat. *UNIX*, *IPv4*, *IPv6*, *NetLink* and *KCAPI* sockets)
. *Lock sandboxing* (uses *Landlock LSM*)
. *Crypt sandboxing* (Transparent File Encryption with *AES-CTR*)
. *Proxy sandboxing* (SOCKS proxy forwarding with network namespace isolation, defaults to *TOR*)
. *Memory sandboxing*
. *PID sandboxing* (simpler alternatives to *Control Groups*)
. *Ghost mode* (similar to Seccomp Level 1 aka Strict Mode)


Sandboxing may be *on* or *off*:

|[ *off*
:< Sandboxing is off, none of the relevant system calls are checked and all access is granted.
|[ *on*
:< Sandboxing defaults to deny, allowlists and denylists can be used to refine access rights.

In addition, there are filters for every sandboxing to prevent Syd
from reporting an access violation. Note, access is still denied in such
cases.

## Read Sandboxing

This sandboxing checks certain system calls for filesystem read access. If a
system call tries to read a path, this attempt is reported and the system call
is denied. See the section called *Write Sandboxing* for more information on how
to customise this behaviour.

List of filtered system calls are: _open_(2), _openat_(2), _openat2_(2),
_getrandom_(2), and _memfd_create_(2).

- _getrandom_(2) with the flag "GRND_RANDOM" is treated equivalent to \
"open(/dev/random, O_RDONLY)".
- _getrandom_(2) without the flag "GRND_RANDOM" is treated equivalent to \
"open(/dev/urandom, O_RDONLY)".
- _memfd_create_(2) name argument is prepended with "/memfd:" before access \
check. Use e.g. "deny/read+/memfd:\*" to deny access to memory file \
descriptors globally.

## Write Sandboxing

This sandboxing checks certain system calls for filesystem write access. If a system
call tries to write, modify or change attributes of a path, this attempt is reported
in system log and the system call is denied. There are two ways to customise this
behaviour. Syd may be configured to "allow" some _glob_(3p) patterns. If the path
argument of the system call which is subject to be modified matches a pattern in the
list of allowed _glob_(3p) patterns, this attempt is not denied. If, however it
matches a pattern in the list of "deny" _glob_(3p) patterns the attempt is denied
(_last matching pattern wins_). Additionally, Syd may be configured to
"filter" some _glob_(3p) patterns. In this case a match will prevent Syd from
reporting a warning about the access violation, the system call is still denied
though.

List of filtered system calls are: _chmod_(2), _fchmod_(2),
_fchmodat_(2), _chown_(2), _chown32_(2), _fchown_(2), _fchown_(32),
_lchown_(2), _lchown32_(2), _fchownat_(2), _open_(2), _openat_(2),
_openat2_(2), _creat_(2), _mkdir_(2), _mkdirat_(2), _mknod_(2),
_mknodat_(2), _rmdir_(2), _truncate_(2), _truncate64_(2), _utime_(2),
_utimes_(2), _utimensat_(2), _futimesat_(2), _unlink_(2), _unlinkat_(2),
_link_(2), _linkat_(2), _rename_(2), _renameat_(2), _renameat2_(2),
_symlink_(2), _symlinkat_(2), _setxattr_(2), _fsetxattr_(2),
_lsetxattr_(2), _removexattr_(2), _fremovexattr_(2), _lremovexattr_(2),
and _memfd_create_(2).

- _memfd_create_(2) name argument is prepended with "/memfd:" before access \
check. Use e.g. "deny/write+/memfd:\*" to deny access to memory file \
descriptors globally.

## Stat Sandboxing

This sandboxing checks certain system calls for file metadata access.
This can be one of listing a directory, changing into a directory, or using a
_stat_(2) system call to query file metadata. This sandboxing type may be used to
effectively _hide files and directories_ from the sandbox process.

List of filtered system calls are: _access_(2), _faccessat_(2), _faccessat2_(2),
_chdir_(2), _fchdir_(2), _getdents_(2), _getdents64_(2), _stat_(2), _fstat_(2),
_lstat_(2), _statx_(2), _newfstatat_(2), _readlink_(2), _readlinkat_(2),
_getxattr_(2), _lgetxattr_(2), _fgetxattr_(2), _listxattr_(2), _flistxattr_(2),
_llistxattr_(2), _statfs_(2), _statfs64_(2), _fstatfs_(2), _fstatfs64_(2),
_fanotify_mark_(2), and _inotify_add_watch_(2).

## Exec Sandboxing

This sandboxing denies _execve_(2), and _execveat_(2) calls in case the
path argument does not match one of the allowed _glob_(3p) patterns.
Note, all _exec_(2) family functions are sandboxed because these
functions are just wrappers of either one of _execve_(2) or
_execveat_(2) system calls. For scripts, access check is done for both
the script and the shebang program . This is also true for Force
Sandboxing.

As of version 3.16.3, Syd checks the paths of the dynamic libraries
an executable is linked against for exec access as well. This only
works for ELF files.

As of version 3.21.1, Syd checks the _memfd_create_(2) system call for
exec access as well because these files are created with mode 777.
_memfd_create_(2) name argument is prepended with "/memfd:" before
access check. Use e.g. "deny/exec+/memfd:\*" to deny access to memory
file descriptors globally.

As of verison 3.21.2, Syd seals memory file descriptors as
non-executable by default. Therefore memory file descriptors are not
checked for Exec access unless the option `trace/allow_unsafe_memfd:1`
is set.

As of version 3.21.3, Syd hooks into _mmap_(2), and _mmap2_(2) system
calls and checks the file descriptor for Exec access when the memory
protection mode includes "PROT_EXEC" and flags does not include
"MAP_ANONYMOUS" which typically indicates a _dlopen_(3). Therefore
libraries dynamically loaded at runtime are checked for Exec access as
well.

## Ioctl Sandboxing

This sandboxing checks the _ioctl_(2) system call for filesystem access.
In addition, unwanted _ioctl_(2) requests can be denylisted by adding
them to the denylist with "ioctl/deny+".

## SegvGuard

As of version syd-3.16.3, Syd has a simple implementation of SegvGuard.
The implementation is inspired by that of HardenedBSD with identical
defaults: If a sandbox process produces a core dump for
"segvguard/maxcrashes" times (defaults to 5), in a period of
"segvguard/expiry" seconds (defaults to 2 minutes), subsequent attempts
to execute the same executable is denied for "segvguard/suspension"
seconds (defaults to 10 minutes). SegvGuard can be disabled either by
setting "segvguard/expiry:0" or by setting "trace/allow_unsafe_exec:1".
The former setting works at runtime whereas the latter must be passed
on startup. SegvGuard is guarded by the exec-TOCTOU mitigator.

Check out the following links for further information on SegvGuard:

- http://en.wikibooks.org/wiki/Grsecurity/Appendix/Grsecurity_and_PaX_Configuration_Options#Deter_exploit_bruteforcing
- http://en.wikibooks.org/wiki/Grsecurity/Appendix/Grsecurity_and_PaX_Configuration_Options#Active_kernel_exploit_response
- http://phrack.org/archives/issues/59/9.txt
- http://phrack.org/archives/issues/58/4.txt
- https://github.com/HardenedBSD/hardenedBSD/wiki/segvguard2-ideas---brainstorm

## Force Sandboxing

Force Sandboxing enhances system security by scrutinizing the path
provided to _execve_(2) and _execveat_(2) system calls, comparing them
against a predefined Integrity Force map -- a registry of
path-to-checksum correlations. Upon invocation of these calls, the
sandbox computes the checksum of the target binary and cross-references
it with the map. Discrepancies trigger rule-defined actions: execution
might proceed with a logged warning, or culminate in the termination of
the process in violation. This mechanism allows for rigorous enforcement
of binary integrity, echoing the preventative ethos of HardenedBSD's
Integriforce and NetBSD's Veriexec by proactively mitigating
unauthorised code execution, albeit with a unique emphasis on flexible,
user-defined consequence management ranging from permissive alerts to
stringent execution blocks.

Distinguishing itself through user-centric customization, Force Sandboxing
offers a versatile approach to execution integrity. Administrators can tailor
the sandbox's response to checksum mismatches -- kill, or warn -- thereby
balancing security needs with operational flexibility. This adaptability,
combined with tools like _syd-sha_(1) for checksum calculation and _syd-path_(1)
for rule creation, positions Force Sandboxing as a powerful ally in the
preservation of system integrity. See _force_ command in _syd_(2) manual
page on how to add/remove entries to/from the Integrity Force map.

As of version 3.16.3, Syd checks the paths of the dynamic libraries an
executable is linked against for force access as well. This only works
for ELF files.

As of version 3.21.3, Syd hooks into _mmap_(2), and _mmap2_(2) system
calls and checks the file descriptor for Force access when the memory
protection mode includes "PROT_EXEC" and flags does not include
"MAP_ANONYMOUS" which typically indicates a _dlopen_(3). Therefore
libraries dynamically loaded at runtime are checked for Force access as
well.

## TPE sandboxing

As of version 3.21.0, Syd introduces Trusted Path Execution (TPE)
sandboxing, which restricts the execution of binaries to ensure they
come from trusted directories. The intention is to make privilege
escalation harder when an account restricted by TPE is compromised as
the attacker won't be able to execute custom binaries which are not in
the trusted path. Trusted directories are defined as those that meet the
following criteria:

. Not writable by group or others.
. Optionally owned by root (controlled by the "tpe/root_owned" setting).
. Optionally owned by the current user or root (controlled by the "tpe/user_owned" setting).

If the directories do not meet these criteria, the execution is denied
with an "EACCES" error, and optionally, the offending process can be
terminated with "SIGKILL" using the "tpe/kill" setting. This mechanism
ensures that only binaries from secure, trusted paths can be executed,
enhancing security by preventing unauthorized code execution. TPE
sandboxing operates by checking the parent directories of executables at
two stages:

. During the system call entry of _execve_(2) and _execveat_(2).
. On _ptrace_(2) exec event to include all dynamic libraries loaded by the executable.

TPE can be configured to apply to a specific user group. By default, TPE
applies to all users. However, administrators can specify an untrusted
GID with the "tpe/gid" setting, restricting TPE only to users in that
group. Additionally, TPE can negate GID logic with the "tpe/negate"
setting, making the specified group trusted and exempt from TPE.

Syd's TPE implementation is based on HardenedBSD's which is inspired
by GrSecurity's TPE. Check out the following links for more information:

- http://phrack.org/issues/52/6.html#article
- http://phrack.org/issues/53/8.html#article
- https://wiki.gentoo.org/wiki/Hardened/Grsecurity_Trusted_Path_Execution

## Network Sandboxing

This sandboxing exposes a way to prevent unwanted network calls. The filtered
system calls are: _bind_(2), _connect_(2), _sendto_(2), _sendmsg_(2), and
_sendmmsg_(2). To increase usability, these system calls are filtered in two
groups: "bind" and "connect". _bind_(2) belongs to the "bind" group, whereas the
other system calls belong to the "connect" group. See the section called *Write
Sandboxing* for more information on how to customise this behaviour.

## Lock Sandboxing

This sandboxing utilises the *Landlock Linux Security Module* for simple
unprivileged access control. This sandboxing type is not dynamic and is applied
at the kernel level on startup. The sandboxing may be turned on with the
"sandbox/lock:on" sandbox command, and _read-only_, and _read-write_ allowlists
can be populated using the sandbox commands "allow/lock/read+/path" and
"allow/lock/write+/path". Note the Syd process is also included in this
sandbox for added security such that a compromised Syd process is still stuck
inside the *Landlock* sandbox.

## Crypt Sandboxing

This sandboxing provides transparent file encryption using *AES-CTR*,
ensuring secure data handling without manual encryption steps. When
"sandbox/crypt:on" is set, files matching the _glob_(3) patterns
specified by "crypt+" are encrypted on write and decrypted on read.
Configuration includes specifying a 256-bit AES key in hexadecimal using
"crypt/key". To avoid including the key into the configuration file, the
user may set the key using an environment variable and then specify this
environment variable, e.g: "crypt/key:${SYD_KEY}". The user _must_ use
an environment variable name that starts with the prefix "SYD_" but does
not start with the prefix "SYD_TEST_" as such environment variables
don't leak into the sandbox process. Similarly the user should refrain
from using any environment variable specified under ENVIRONMENT section
of the _syd_(1) manual page. Encryption operates via Linux kernel
cryptography API sockets, using zero-copy techniques with _splice_(2) to
avoid unencrypted data in memory. The setup socket is created on
startup, the key is set using _setsockopt_(2), and then *the key is
wiped from memory before sandboxing starts. Syd also ensures that the
key and IV never swaps out.* IV uniqueness is ensured by generating a
random IV using /dev/random per file. Per-file IV is saved into the
"user.crypt.syd.iv" extended attribute of the file. This ensures
security by preventing IV reuse. By default, decryption occurs in a
memory file descriptor to prevent tampering, which limits practicality
for large files due to memory constraints. User may specify a secure
temporary backing directory with "crypt/tmp" to solve this.
Ideally this directory should be on encrypted storage. During
transparent file decryption, Syd creates files in this directory with
the "O_TMPFILE" flag. No effort is made to secure the contents of these
plaintext files other than the kernel enforcing their deletion on Syd
exit. File locks are set before attempting to get/set extended
attributes to ensure security. When a file is locked for encryption or
decryption, any subsequent attempts to access the file with read/write
operations will block ensuring exclusive access during these processes.
Concurrent read access is allowed, but write access will block other
reads and writes. Only regular files will be encrypted. Append to
encrypted files are handled efficiently by reencrypting the last partial
block. These blocks that leak into Syd's memory are never swapped out
and zeroed on drop. The extended attribute "user.syd.crypt.api" mark
files as encrypted and its value must match the current Syd API which at
the moment is 3. If this extended attribute can't be set on a file, it
won't be encrypted. The attribute is set at the beginning of the
encryption process to allow partial decryption of interrupted encryption
results. Syd ensures extended attributes that start with the prefix
"user.syd." can't be listed or tampered by the sandbox process. Compared
to GSWTK's dbfencrypt, Crypt sandboxing avoids TOCTOU vulnerabilities
and encryption weaknesses by utilizing AES-CTR and robust setup steps,
providing a more secure and streamlined encryption process.

Crypt sandboxing employs the *AES-CTR* algorithm, a secure and efficient
symmetric key encryption method suitable for various applications. It operates
as a stream cipher (skcipher) with a block size of 1 byte, allowing data to be
encrypted in a byte-by-byte manner. The algorithm uses a fixed key size of 32
bytes (256 bits) by default, providing robust security, and a fixed
initialization vector (IV) size of 16 bytes to ensure randomness and uniqueness
in each encryption operation. Processing data in byte-sized chunks, the
algorithm maintains a consistent walk size of 16 bytes for traversal and
operations, ensuring seamless encryption and decryption processes. This
configuration, with its secure default key size, significantly enhances
security, preventing common encryption weaknesses and supporting efficient,
transparent file encryption within the sandbox environment. *CTR* is infinitely
parallelizable because each block in the stream can be encrypted independently.
This allows for encryption and decryption processes to be split across multiple
processors, significantly increasing throughput. With hardware support such as
*AES-NI* CPU instructions, speeds can easily exceed a gigabyte per second.

As of version 3.21.2, Syd opens memory file descriptors with the flag
"MFD_NOEXEC_SEAL" during transparent decryption to ensure the memfds are
non-executable and can't ever be marked executable. This ensures
security as otherwise transparent decryption can be used to bypass Exec,
Force and TPE sandboxing. Notably, this flag requires Linux-6.3 or
newer. On older kernels, a backing directory must be specified with
"crypt/tmp" for transparent decryption to work. Attempt to use
transparent decryption without a backing directory on older kernels will
fail with the errno "EOPNOTSUPP".

Limitations:

- *Incompatibility with O_RDWR Operations*: The current implementation
  treats "O_RDWR" (aka read-write) operations as "O_WRONLY" (aka
  write-only), rendering simultaneous read and write operations on a file
  impossible. Specifically, any attempt to seek during these operations
  will result in an "ESPIPE" error, indicating an invalid seek. This
  design choice prioritizes secure, exclusive access during encryption and
  decryption but sacrifices the ability to perform concurrent read and
  write operations on the same file.
- *IV Save in Xattrs*: If a file is copied and the extended attributes
  (xattrs) are stripped, the IV is lost, resulting in data loss. This
  makes handling file backups and transfers potentially dangerous if
  xattrs are not preserved. Users must ensure xattrs are retained during
  file operations to avoid permanent data loss.
- *Lack of Integrity Checking*: The current implementation does not
  provide integrity checking for encrypted files, meaning that
  unauthorized modifications or corruption of data might go undetected.
  Future updates plan to switch from AES-CTR to AES-GCM, which will
  include built-in integrity checking. AES-GCM (Galois/Counter Mode) not
  only encrypts data but also provides authentication, ensuring that any
  unauthorized changes to the data are detected. This enhancement will
  significantly improve data security by protecting against tampering and
  ensuring data integrity.
- *Large files* are not handled efficiently during decryption by default due to
  usage of in-memory files, specify a secure temporary backing directory with
  "crypt/tmp:/path" to solve this.

Utilities:
- _syd-aes_(1): Encrypts/decrypts files akin to openssl-enc.
- _syd-cp_(1): Efficient file copying under Syd using _splice_(2).
- _syd-key_(1): Generates random AES-CTR key using "/dev/random".

## Proxy Sandboxing

As of version 3.22.0, Proxy sandboxing in Syd confines network communication
exclusively through a designated SOCKS proxy, enforced by the helper utility
_syd-tor_(1). Configured at startup with "sandbox/proxy:on", this type implies
the use of "unshare/net:1", isolating network namespaces to prevent direct
network access. Traffic is forwarded from a specified local port
(proxy/port:9050) to an external address and port (proxy/ext/addr:127.0.0.1,
proxy/ext/port:9050). This setup ensures all network interactions route through
the proxy, leveraging zero-copy data transfers and edge-triggered _epoll_(7) for
efficient event handling. The implementation enhances security by employing
seccomp and Landlock for additional confinement, preventing unauthorized network
access and ensuring strict adherence to the defined network path. This approach
minimizes the risk of proxy bypasses and maintains the integrity of the network
isolation.

## Memory Sandboxing

This sandboxing handles the system calls _brk_(2), _mmap_(2), _mmap2_(2),
and _mremap_(2) and checks the per-process memory usage on each memory
allocation request. If the memory usage reaches the maximum value
defined by *mem/max*, the system call is denied with *ENOMEM*. Moreover
the virtual memory size can be limited using *mem/vm_max*. If the limit
is reached on the entry of any of the respective system calls, the
system call is denied with *ENOMEM*. If *mem/kill* is set to
true, the signal *SIGKILL* is delivered to the offending process.

The per-process memory usage is a fair estimate calculated using the file
*/proc/pid/smaps* summing the following fields together:

. *"Pss" (Proportional Set Size)* is similar to *Rss*, but \
accounts for shared memory more accurately by dividing it \
among the processes that share it.
. *"Private_Dirty"* represents the private memory that has \
been modified (dirty).
. *"Shared_Dirty"* represents the shared memory that has \
been modified.

Note, *"Rss" (Resident Set Size)* is the portion of memory occupied by a
process that is held in RAM.

Note, _memory sandboxing is not an alternative to cgroups_. You should use
_cgroups_(7) when you can instead. This is meant for more constrained
environments where _cgroups_(7) is not supported or not available due to missing
permissions or other similar restrictions.

Warning, _this is not a strict limit_, Syd is reliant on the allocation
granularity of the processes in the sandbox. Programs will be able to overstep
the limit at least by amount of memory they allocate at once.

## PID sandboxing

This sandboxing handles the system calls _fork_(2), _vfork_(2), _clone_(2), and
_clone3_(2) and checks the total number of tasks running on the system on each
process creation request. If the count reaches the maximum value defined by
*pid/max*, the system call is denied with "EAGAIN". If "pid/kill" is set to
true, the signal "SIGKILL" is delivered to the offending process.

This sandboxing type is best coupled with a pid namespace using "unshare/pid".
In this mode, Syd will check the number of running tasks _in the current
namespace only._

Note, _PID sandboxing is not an alternative to cgroups_. You should use
_cgroups_(7) when you can instead. This is meant for more constrained
environments where _cgroups_(7) is not supported or not available due to missing
permissions or other similar restrictions.

## Ghost mode

Ghost Mode, introduced in Syd version 3.20.0, is the final sandboxing
type, closely resembling Seccomp Level 1, also known as Strict Mode.
This mode enhances security by allowing a process to transition to a
highly restrictive state after completing its initial setup. When a
sandboxed process is ready for this higher level of confinement, it
invokes Ghost Mode by executing the "ghost" command using the virtual
stat API, e.g. with stat("/dev/syd/ghost", NULL). This system call
invocation is going to fail with the errno "EOWNERDEAD" if Ghost mode is
successfully initiated. Upon receiving this command, Syd closes the
_seccomp_(2) notify file descriptor. This action elevates all previously
hooked system calls to a kernel-level deny with the "ENOSYS" errno,
effectively making them unavailable. The transition to Ghost Mode is
irreversible; once the file descriptor is closed, the process is locked
into this restricted state. This mechanism ensures that the sandboxed
process can only perform a very limited set of operations, akin to those
allowed in Seccomp Level 1, thus significantly reducing its potential
attack surface. Ghost Mode provides a robust security measure by denying
all but the most essential system calls, which is crucial for
applications that require maximum isolation and security after their
initial configuration phase.

The mode is aptly named "Ghost" because, upon closing the _seccomp_(2)
notify file descriptor, the sandboxed process effectively detaches from
Syd and becomes independent, much like a ghost. From this point onwards,
Syd no longer intervenes or manages the process's system calls but
merely waits for the process to exit. This detachment underscores the
finality and isolation of the Ghost Mode, ensuring that the process
operates in a secure, tightly confined environment without further
interaction from Syd. This mechanism is particularly useful for
processes that require maximum security and minimal system call exposure
after their initial configuration phase, providing a robust layer of
protection against various exploits and vulnerabilities.

# PROFILES

|[ *container*
:< Enables Linux namespaces.
|[ *immutable*
:< Enables Linux namespaces and remounts the following directories _read only_ in the new mount namespace:
|[
:< */etc*, */home*, */media*, */mnt*, */opt*, */srv*, and */usr*.
|[
:< Further mount options such as *nodev*, *noexec*, *nosuid*, and *noatime* are also applied as necessary.
|[
:< In addition, */dev/shm* and */tmp* are mounted private and kernel filesystems are masked.
|[
:< See *syd-cat immutable* for the full list of mount options.
|[ *privileged*
:< Do not drop Linux capabilities at startup. Used to construct privileged containers.
|[
:< You may use the short name *priv*, or even just *p*, for this profile.
|[ *landlock*
:< Enables *LandLock* and allows system directories for *Lock Sandboxing*.
|[ *paludis*
:< Used by the *Paludis* package mangler.
|[ *noipv4*
:< Disables *IPv4* connectivity.
|[ *noipv6*
:< Disables *IPv6* connectivity.
|[ *silent*
:< Silences all access violations.
|[ *lib*
:< *libsyd* helper profile. Turns all sandboxing off.
|[
:< Useful to configure Syd in the application using *libsyd*.
|[ *user*
:< Allows user-specific directories, and connections, and parses the following files if they exist:
|[
:< */etc/user.syd-3*
|[
:< *~/.user.syd-3*

## Stacking Profiles

It is possible to stack multiple profiles to configure a more restricted
sandbox. Remember the order you stack the profiles matter, _the last matching
rule wins_. Below are some examples:

- syd -puser -pimmutable -mroot:/mnt/gnu ...
- syd -ppaludis -plandlock -mallow/lock/write+/var/tmp ...

## Login shell and the User Profile

When invoked without arguments, */bin/sh* is executed under Syd with the
*user* profile as a login shell, use *SYD_SH* environment variable to override
the shell to execute.

# SECURITY

Syd stands out for its ability to operate without requiring elevated
privileges, eliminating the need for root access. This feature
significantly simplifies setup and usage. Users benefit from the
capability to dynamically configure the sandbox from within, with
options to secure it further as needed. However, the implementation
exercises caution, especially in mitigating *time-of-check to time-of-use (TOCTOU)*
vulnerabilities. Syd employs *SECCOMP_USER_NOTIF_FLAG_CONTINUE* with
considerable restraint, primarily within the hooks for _exec_(2) and _chdir_(2)
system calls. These calls proceed directly in the sandbox process due to the
inability to emulate them fully within _seccomp_(2)'s constraints, this may
change in the future when the Linux kernel gains support to allow seccomp to
emulate them. For detailed information on specific limitations and their
security implications, refer to the *BUGS* section.

*Tip*: To take a quick peek at the seccomp filters applied by Syd under
various different configurations, use *syd <flags...> -Epfc*.

## Accessing remote process memory

Syd denies various system calls which can access remote process memory such
as _ptrace_(2) and _process_vm_writev_(2) and common sandboxing profiles such as
"paludis" and "user" disallow write access to the "/proc/pid/mem" file. This
makes "TOCTOU" attack vectors harder to realise. See "syd-ls deny" to see
the full list of denied system calls.

## Enhanced Handling of PTRACE_TRACEME

As of version 3.16.3, Syd introduced a new feature for managing the
"PTRACE_TRACEME" operation, aimed at improving stealth against detection.
Traditionally, "PTRACE_TRACEME" is the only _ptrace_(2) operation allowed by a
tracee, which makes it a common target for detection of ptracers. By converting
"PTRACE_TRACEME" into a no-operation (no-op) that always succeeds, Syd aims to
subtly prevent simple detection methods that rely on this operation.
Additionally, other _ptrace_(2) operations are modified to return an EPERM
(Permission Denied) error instead of ENOSYS (Function Not Implemented), which
helps reduce the likelihood of the sandbox being detected through these errors.
This approach enhances the discreetness of Syd's operation by mitigating
straightforward detection tactics used by monitored processes.

As of version 3.19.0, Syd extends this mitigation and turns the system
call _ptrace_(2) into a no-op. Again, this provides a best-effort
mitigation against using requests such as "PTRACE_ATTACH" or
"PTRACE_SEIZE" to detect a ptracer.

## Hardened procfs

As of version 3.14.5, Syd has introduced a significant enhancement to its
sandbox tool, adopting a process hardening approach similar to that utilised by
HardenedBSD's handling of procfs/linprocfs. This new security measure
specifically targets and restricts access to a set of critical paths within the
process file system, including the files "auxv", "environ", "maps", "mem",
"pagemap", "personality", "stack", "syscall", and "wchan" under the directories
/proc/$pid and /proc/$pid/task/$tid. Access to these paths is denied by
default, regardless of the process or thread identifier (PID/TID) involved or
the state of sandboxing. This strict policy of unconditional denial is designed
to further secure the system by preventing unauthorised access to sensitive
process information, thereby enhancing the overall integrity and security
posture of the environment. This addition complements existing protections
against remote process memory access. Together, these measures significantly
mitigate the risk of _Time-of-Check to Time-of-Use (TOCTOU)_ attack vectors,
underscoring Syd's commitment to providing a robust security framework for its
users. As of version 3.15.1, Syd further hardens the procfs restrictions by
denying access to the files /proc/cmdline, /proc/config.gz, /proc/dma,
/proc/interrupts, /proc/iomem, /proc/ioports, /proc/kallsyms,
/proc/kcore, /proc/key-users, /proc/keys, /proc/modules, /proc/mtrr,
/proc/partitions, /proc/softirqs, /proc/sysrq-trigger, /proc/version,
/proc/vmallocinfo and directories /proc/fs, /proc/irq, and /proc/tty
regardless of the sandboxing mode.

## Restricted Device Access

As of version 3.15.1, Syd introduces a critical security feature that
significantly bolsters the integrity and confidentiality of the running kernel
and system devices. Building upon the foundational principles of the hardened
procfs feature, this update extends protection mechanisms to critical device
files, specifically targeting "/dev/mem", "/dev/kmem", "/dev/port", and
"/dev/cpu/\*". This enhancement prevents both reading from and writing to these
sensitive areas, effectively safeguarding the system against unauthorised
modifications or leaks of kernel memory and preventing misuse of system ports
and CPU-specific registers.

## Changing UID/GID and groups

Since 3.1.9, the set-id family system calls, _setuid_(2), _setgid_(2) et
al., are no-ops under Syd. Since 3.8.6, the Syd process applies a
seccomp filter to itself as well to get the same behaviour. This means
even a compromised Syd process cannot elevate privileges using these
system calls. Making these system calls no-ops rather than denying them
provides easy compatibility with daemons that e.g. drop user:group to
"nobody:nogroup". See "syd-ls setid" to see the full list of system
calls in this group.

Since version 3.16.8, *safesetid* has been introduced in Syd, enhancing
the management of UID/GID transitions by replacing the previous
"trace/allow_unsafe_setid" configuration. This feature enables
finer-grained control by allowing administrators to explicitly specify
permissible transitions for UID and GID changes, thus tightening
security constraints around process privilege management. *safesetid*
can be enabled by "trace/allow_safe_setid:1" and it works by allowing
predefined UID and GID transitions that are explicitly configured using
the "setuid+<source_uid>:<target_uid>" and
"setgid+<source_gid>:<target_gid>" commands in the Syd configuration.
This ensures that transitions can only occur between specified user and
group IDs, and unauthorised privilege escalations are blocked. For
instance, a transition might be allowed from a higher-privileged user to
a less-privileged user but not vice versa, thereby preventing any
escalation of privileges through these system calls.

By default, Syd sets all set-id family system calls to no-ops, reinforcing its
security-first approach. This default behavior means that without specific
configuration via "trace/allow_safe_setid:true", even a compromised Syd process
cannot use these system calls to change its user or group ID in an unauthorised
manner. Enabling "allow_safe_setid" requires explicit declarations of each
allowed transition to ensure that only specified changes to UID and GID are
possible. This controlled approach maintains the default security posture of
Syd while providing flexibility where necessary through meticulous
configuration.

## Denying TIOCLINUX ioctl

The limitation on the use of the "TIOCLINUX" _ioctl_(2) within secure
environments, similar to the Syd sandbox, is an essential security measure
addressing vulnerabilities specific to Linux terminal operations. The
"TIOCLINUX" _ioctl_(2) command offers various functionalities, including but not
limited to manipulating console settings, changing keyboard modes, and
controlling screen output. While these capabilities can be leveraged for
legitimate system management tasks, they also introduce potential security
risks, particularly in multi-user environments or in the context of sandboxed
applications.

The security concerns surrounding "TIOCLINUX" stem from its ability to alter
terminal behaviors and settings in ways that could be exploited for unauthorised
information disclosure, terminal hijacking, or privilege escalation. For
instance, manipulating the console display could mislead users about the true
nature of the operations being executed, or altering keyboard settings could
capture or inject keystrokes.

In summary, the restriction on "TIOCLINUX" within secure environments is a vital
security strategy, addressing the complex risks associated with direct terminal
manipulation capabilities. This precaution is in keeping with the broader
security community's efforts to mitigate known vulnerabilities and enhance the
security posture of systems handling sensitive processes and data.

## Denying TIOCSTI ioctl

The restriction on the use of the "TIOCSTI" _ioctl_(2) within the Syd
sandbox addresses a significant security vulnerability associated with
terminal input injection. The "TIOCSTI" _ioctl_(2) allows a byte to be
inserted into the terminal input queue, effectively simulating keyboard
input. This capability, while potentially useful for legitimate
purposes, poses _a substantial security risk_, especially in scenarios
where a process might retain access to a terminal beyond its intended
lifespan. Malicious use of this _ioctl_(2) can lead to the injection of
commands that execute with the privileges of the terminal's owning
process, thereby breaching the security boundaries intended by user
permissions and process isolation mechanisms. The concern over "TIOCSTI"
is well-documented in the security community. For example, OpenBSD has
taken measures to mitigate the risk by disabling the "TIOCSTI"
_ioctl_(2), reflecting its stance on the _ioctl_(2) as _one of the most
dangerous_ due to its potential for abuse in command injection attacks.
The decision to disable or restrict "TIOCSTI" in various Unix-like
operating systems underscores the _ioctl_(2)'s inherent security
implications, particularly in the context of privilege escalation and
the execution of unauthorised commands within a secured environment.

In summary, the restriction on "TIOCSTI" within Syd is a critical
security measure that prevents a class of vulnerabilities centered
around terminal input injection, safeguarding against unauthorised
command execution and privilege escalation. This precaution aligns with
broader security best practices and mitigations adopted by the security
community to address known risks associated with terminal handling and
process isolation.

## Denying PR_SET_MM prctl

The "PR_SET_MM" _prctl_(2) call allows processes with the "CAP_SYS_RESOURCE"
capability to adjust their memory map descriptors, facilitating operations like
self-modifying code by enabling dynamic changes to the process's memory layout.
For enhanced security, especially in constrained environments like Syd, this
capability is restricted to prevent unauthorised memory manipulations that could
lead to vulnerabilities such as code injection or unauthorised code execution.
Notably, Syd proactively drops "CAP_SYS_RESOURCE" among other capabilities at
startup to minimise security risks. This action is part of Syd's broader
security strategy to limit potential attack vectors by restricting process
capabilities. Users can query "syd-ls drop" for a comprehensive list of
capabilities that Syd relinquishes upon initialization, reinforcing the
sandbox's commitment to a secure execution environment by adhering to the
principle of least privilege.

## Restricting prctl option space and trace/allow_unsafe_prctl

Syd meticulously confines the scope of permissible _prctl_(2) operations to
enhance security within its sandbox environment. By limiting available
_prctl_(2) options to a specific set, including but not limited to
"PR_SET_PDEATHSIG", "PR_GET_DUMPABLE", "PR_SET_NO_NEW_PRIVS", and
"PR_SET_SECCOMP", Syd ensures that only necessary process control
functionalities are accessible, thereby reducing the risk of exploitation
through less scrutinised _prctl_(2) calls. This constraint is pivotal in
preventing potential security vulnerabilities associated with broader _prctl_(2)
access, such as unauthorised privilege escalations or manipulations of process
execution states. However, recognizing the need for flexibility in certain
scenarios, Syd offers the option to lift these restrictions through the
"trace/allow_unsafe_prctl:1" setting. This capability allows for a tailored
security posture, where users can opt for a more permissive _prctl_(2)
environment if required by their specific use case, while still maintaining
awareness of the increased security risks involved.

## Restricting io_uring interface and trace/allow_unsafe_uring

The _io_uring_(7) interface can be used to _bypass path sandboxing_. By default,
Syd restricts _io_uring_(7) operations due to their ability to perform system
calls that could undermine the sandbox's security controls, particularly those
designed to limit file access and modify file permissions. The setting,
"trace/allow_unsafe_uring", when enabled, relaxes these restrictions, allowing
_io_uring_(7) operations to proceed unimpeded. While this can significantly
enhance I/O performance for applications that rely on _io_uring_(7) for
efficient asynchronous operations, it requires careful consideration of the
security implications, ensuring that its use does not inadvertently compromise
the sandboxed application's security posture. Use "syd-ls uring" to see the full
list of system calls that belong to the io_uring interface.

## Restricting creation of device special files

Since version 3.1.12, Syd has enhanced its security model by disallowing
the creation of device special files through the _mknod_(2) and
_mknodat_(2) system calls. This decision is rooted in mitigating
potential security vulnerabilities, as device special files could be
exploited to circumvent established path-based access controls within
the sandbox environment. These files, which include character and block
devices, can provide direct access to hardware components or facilitate
interactions with kernel modules that could lead to unauthorised actions
or data exposure. By restricting their creation, Syd significantly
reduces the risk of such exploit paths, reinforcing the integrity and
security of the sandboxed applications. This measure ensures that only
predefined types of files -- such as FIFOs, regular files, and sockets --
are permissible, aligning with the principle of least privilege by
limiting file system operations to those deemed safe within the
sandbox's context.

## Sharing Pid namespace with signal protections

Since version 3.6.7, Syd has introduced a crucial security feature that
prevents sandboxed processes from sending signals to the Syd process or
any of its threads. This protection is implemented by hooking and
monitoring system calls related to signal operations, including
_kill_(2), _tkill_(2), _tgkill_(2), and _pidfd_open_(2). When a
sandboxed process attempts to send a signal to Syd or its threads, these
system calls are intercepted, and the operation is denied at the seccomp
level with an "EACCES" error. This measure ensures that Syd maintains
control over the execution and management of sandboxed processes,
safeguarding against interruptions or unauthorised interactions that
could compromise the security or stability of the sandbox environment.
This security mechanism is part of Syd's broader strategy to share the
same root, private proc, and mount namespaces with the sandboxed
process, facilitating secure and simple system call emulation. By making
Syd and its threads immune to signals from sandboxed processes, the
integrity and isolation of the sandboxed environment are significantly
enhanced, preventing potential exploitation scenarios where sandboxed
processes could disrupt the operation of the sandbox manager or
interfere with other sandboxed processes.

Note, as of version 3.18.10, Syd disables the deprecated _tkill_(2)
system call by default for safety against PID-recycling. For
compatibility, Syd turns this system call into a no-op, a call that
always succeeds without performing any action. The option
"trace/allow_unsafe_kill:1" may be used to relax this restriction.

## System Call Handler Priority and Resource Management

Since version 3.8.1, Syd has been implementing strategies to ensure the
smooth operation of the host system while managing security through its
sandboxing mechanism. It sets the _nice_(2) value of its system call
handler threads to "19", ensuring these threads operate at _the lowest
priority_ to minimise CPU starvation for other critical processes. This
approach prioritises system stability and fair CPU resource
distribution, enabling Syd to handle numerous system calls without
compromising the host's performance and responsiveness.

Enhancing this strategy, Syd introduced further adjustments in versions
3.8.6 and 3.9.7 to address I/O and CPU resource management more
comprehensively. From version 3.8.6, it sets the I/O priority of the
system call handler threads to "idle", ensuring that I/O operations do
not monopolise resources and lead to I/O starvation for other processes.
Similarly, from version 3.9.7, it adjusts the CPU scheduling priority of
these threads to "idle", further safeguarding against CPU starvation.
These measures collectively ensure that Syd maintains optimal
performance and system responsiveness while securely sandboxing
applications, striking a balance between security enforcement and
efficient system resource utilization.

## Simplified Filesystem Access for Legacy Daemons

Since version 3.12.0, Syd has adopted a nuanced approach to managing
filesystem access controls, specifically concerning the _chroot_(2) and
_pivot_root_(2) system calls. In Syd, these calls are treated as
no-operations (no-ops), meaning they _report success without performing
any action_. This decision is strategically aimed at accommodating older
daemons that invoke "chroot(/var/empty)" as part of their initialization
sequence, typically after completing necessary path accesses. By
rendering _chroot_(2) and _pivot_root_(2) as no-ops at the seccomp-bpf
kernel level, Syd ensures compatibility with such legacy software
without introducing complexities in syscall emulation.

The rationale behind this approach stems from the necessity of
maintaining a unified filesystem root between the sandbox and the
sandboxed processes to simplify syscall emulation. However, certain
daemons, like OpenNTPD, traditionally employ _chroot_(2) to a minimal
directory (e.g., "/var/empty") as a security measure. Syd's no-op policy
for these system calls allows it to host these daemons without
necessitating alterations to their source code or operational
logic, thus broadening the utility of Syd in securing legacy
applications without sacrificing operational integrity or
complicating the sandbox's implementation.

## Streamlining File Synchronization Calls

Since version 3.8.8, Syd has rendered the _sync_(2) and _syncfs_(2)
system calls as no-operations (no-ops), ensuring they report success
without executing any underlying functionality. This adjustment is
designed to streamline operations within the sandboxed environment,
bypassing the need for these file synchronization actions that could
otherwise impact performance or complicate the sandbox's control over
file system interactions. By adopting this approach, Syd enhances its
compatibility with applications that issue these calls, without altering
the sandboxed process's behavior or the integrity of file system
management.

## Restricting Resource Limits, Core Dumps, and trace/allow_unsafe_prlimit

Since version 3.9.6, Syd has implemented restrictions on setting process
resource limits and generating core dumps for the sandboxed process,
enhancing the sandbox's security posture. This measure prevents the
sandboxed process from altering its own resource consumption boundaries
or producing core dumps, which could potentially leak sensitive
information or be exploited for bypassing sandbox restrictions. However,
recognizing the need for flexibility in certain use cases, Syd provides
the option to disable these restrictions at startup through the
"trace/allow_unsafe_prlimit:1" setting. This allows administrators to
tailor the sandbox's behavior to specific requirements, balancing
security considerations with functional needs.

## Enhancing Sandbox Security with Landlock

Since version 3.0.1, Syd leverages the *Landlock Linux Security Module
(LSM)* to enforce advanced filesystem sandboxing, significantly
bolstering the security framework within which sandboxed processes
operate. By integrating Landlock, Syd empowers even unprivileged
processes to create secure sandboxes, enabling fine-grained access
control over filesystem operations without requiring elevated
permissions. This approach is instrumental in mitigating the risk of
security breaches stemming from bugs or malicious behaviors in
applications, offering a robust layer of protection by restricting
ambient rights, such as global filesystem or network access. Landlock
operates by allowing processes to self-impose restrictions on their
access to system resources, effectively creating a secure environment
that limits their operation to a specified set of files and directories.
This mechanism is particularly useful for running legacy daemons or
applications that require specific environmental setups, as it allows
for the precise tailoring of access rights, ensuring processes can only
interact with designated parts of the filesystem. For instance,
by setting Landlock rules, Syd can confine a process's
filesystem interactions to read-only or read-write operations on
explicitly allowed paths, thus preventing unauthorised access to
sensitive areas of the system.

Furthermore, the inclusion of the Syd process itself within the
Landlock-enforced sandbox adds an additional layer of security. This
design choice ensures that even if the Syd process were compromised, the
attacker's ability to manipulate the sandboxed environment or access
unauthorised resources would be significantly constrained. This
self-sandboxing feature underscores Syd's commitment to maintaining a
high security standard, offering peace of mind to users by ensuring
comprehensive containment of sandboxed processes.

## Namespace Isolation in Syd

Syd enhances sandbox isolation through meticulous namespace use, starting
from version 3.0.2. Version 3.9.10 marks a pivotal enhancement by restricting
user subnamespace creation, addressing a key path sandboxing bypass
vulnerability. This strategic limitation thwarts sandboxed processes from
altering their namespace environment to access restricted filesystem areas.
Furthermore, since version 3.11.2, Syd maintains process capabilities within
user namespaces, mirroring the _unshare_(1) command's "--keep-caps" behavior.
This ensures sandboxed processes retain necessary operational capabilities,
enhancing security without compromising functionality. Additionally, Syd
utilises the powerful "bind" command within the mount namespace to create
secure, isolated environments by allowing specific filesystem locations to be
remounted with custom attributes, such as "read-only" or "no-exec", providing a
flexible tool for further restricting sandboxed processes' access to the
filesystem.

Syd also introduces enhanced isolation within the mount namespace by
offering options to bind mount temporary directories over "/dev/shm" and
"/tmp", ensuring that sandboxed processes have private instances of these
directories. This prevents inter-process communication through shared memory
and mitigates the risk of temporary file-based attacks, further solidifying the
sandbox's defence mechanisms. Moreover, since version 3.11.2, Syd mounts
the proc filesystem privately with the "hidepid=2" option, enhancing privacy by
concealing process information from unauthorised users.

Syd's "container" and "immutable" profiles exemplify its adaptability,
offering from isolated to highly restrictive environments. The "container"
profile provides a general-purpose sandbox, while the "immutable" profile
enforces stricter controls, such as making essential system directories
read-only, to prevent tampering. This comprehensive approach underlines Syd's
adept use of kernel features for robust sandbox security, ensuring a secure and
controlled execution environment for sandboxed applications. See "syd-cat
container", and "syd-cat immutable" to list the rules in these sandboxing
profiles.

As of version 3.23.0, Syd has further strengthened its security. The
introduction of a time namespace, represented by the "unshare/time:1"
option, allows Syd to reset the boot-time clock, ensuring that the
_uptime_(1) command reports container uptime instead of host uptime.
Moreover, the creation of namespaces, including mount, UTS, IPC, user,
PID, net, cgroup, and time is denied by default to prevent unauthorized
namespace manipulation that could undermine path sandboxing security. To
allow specific namespace types, administrators must explicitly enable
them via the "trace/allow_unsafe_namespace" setting. Another restriction
to note is that the system calls _mount_(2), _mount_setattr_(2),
_umount_(2), and _umount2_(2) are denied by default unless "mount"
namespace is allowed. This change ensures tighter control over process
capabilities and isolation, reinforcing the defense mechanisms against
potential security breaches.

## Restricting environment and trace/allow_unsafe_env

Syd, from version 3.11.1, has implemented measures to clear unsafe
environment variables, such as "LD_PRELOAD", enhancing security by preventing
the manipulation of dynamic linker behavior by sandboxed processes. This action
mitigates risks associated with dynamic linker hijacking, where adversaries may
load malicious shared libraries to execute unauthorised code, potentially
leading to privilege escalation, persistence, or defence evasion. Variables like
"LD_PRELOAD" allow specifying additional shared objects to be loaded before any
others, which could be exploited to override legitimate functions with malicious
ones, thus hijacking the execution flow of a program. To accommodate scenarios
where developers might need to use these variables for legitimate purposes,
Syd allows this security feature to be disabled at startup with
"trace/allow_unsafe_env:1", offering flexibility while maintaining a
strong security posture. This careful balance ensures that sandboxed
applications operate within a tightly controlled environment, significantly
reducing the attack surface and enhancing the overall security framework within
which these applications run. See "syd-ls env" to see the full list of
environment variables that Syd clears from the environment of the sandbox
process.

## Managing Linux Capabilities for Enhanced Security

Since its 3.0.17 release, Syd strategically curtails specific Linux
_capabilities_(7) for sandboxed processes to bolster security. By revoking privileges
such as "CAP_SYS_ADMIN" among others, Syd significantly reduces the risk of
privilege escalation and system compromise. This proactive measure ensures that
even if a sandboxed process is compromised, its ability to perform sensitive
operations is severely limited. The comprehensive list of dropped capabilities,
including but not limited to "CAP_NET_ADMIN", "CAP_SYS_MODULE", and
"CAP_SYS_RAWIO", reflects a meticulous approach to minimizing the attack surface.
See "syd-ls drop" to see the full list of _capabilities_(7) that Syd drops at
startup.

Exceptions to this stringent policy, introduced in version 3.11.1, such
as retaining "CAP_NET_BIND_SERVICE" with "trace/allow_unsafe_bind:1",
"CAP_NET_RAW" with "trace/allow_unsafe_socket:1", "CAP_SYSLOG" with
"trace/allow_unsafe_syslog:1" and "CAP_SYS_TIME" with
"trace/allow_unsafe_time:1", offer a nuanced security model. These
exceptions allow for necessary network, syslog and time adjustments
within the sandbox, providing flexibility without significantly
compromising security.

Since version 3.12.5, Syd allows the user to prevent dropping capabilities at
startup using the command "trace/allow_unsafe_caps:1". This command may be used to
construct privileged containers with Syd.

This balanced strategy of restricting _capabilities_(7), coupled with selective
permissions, exemplifies Syd's commitment to crafting a secure yet functional
sandbox environment. By leveraging the granularity of Linux _capabilities_(7),
Syd offers a robust framework for safeguarding applications against a variety
of threats, underscoring its role as a pivotal tool in the security arsenal of
Linux environments.

## Path Resolution Restriction For Chdir and Open Calls

In Syd version 3.15.1, a configurable security feature is available to
address the risk of directory traversal attacks by restricting the use
of ".." components in path arguments for _chdir_(2), _fchdir_(2),
_open_(2), _openat_(2), _openat2_(2), and _creat_(2) system calls. This
feature is off by default, ensuring broad compatibility and operational
flexibility for a range of applications. When enabled with the
trace/deny_dotdot:1 command, Syd strengthens its defence mechanisms
against unauthorised directory access, echoing the flexibility seen in
FreeBSD’s vfs.lookup_cap_dotdot sysctl. This allows for a nuanced
approach to filesystem security, where administrators can tailor the
sandbox's behavior to match specific security requirements or
operational contexts. By drawing on the security insights of FreeBSD and
HardenedBSD, Syd provides a versatile toolset for managing path
traversal security, adaptable to the unique demands of various
application environments.

## Enhanced Symbolic Link Validation

As of version 3.13.0, Syd enhances security by enforcing stricter
validation on symbolic links within "/proc/pid/fd", "/proc/pid/cwd",
"/proc/pid/exe", and "/proc/pid/root", addressing potential misuse in
container escape scenarios. Specifically, Syd returns an "ELOOP" error
for attempts to resolve these symbolic links if they do not pertain to
the _current process_, akin to implementing "RESOLVE_NO_MAGICLINKS"
behavior of the _openat2_(2) system call. This measure effectively
hardens the sandbox against attacks exploiting these links to access
resources outside the intended confinement, bolstering the isolation
provided by Syd and mitigating common vectors for privilege escalation
and sandbox escape. As of version 3.14.5, Syd keeps intercepting
path system calls even if sandboxing is off making this protection
unconditional.

As of version 3.21.0, Syd enhances security by enforcing stricter
validation on symbolic links within world-writable +t (sticky bit)
directories. Unlike Grsecurity's "GRKERNSEC_LINK" option, which
restricts following symlinks based on ownership within such directories,
Syd simply denies following symlinks in any directory that is
world-writable and has the sticky bit set, without checking ownership.
This measure effectively prevents /tmp symlink attacks by ensuring that
symbolic links in these directories cannot be used to manipulate files
owned by other users, thereby mitigating common and severe security
exploits.

## Memory-Deny-Write-Execute Protections

Syd version 3.14.1 enhances its security framework by implementing
Memory-Deny-Write-Execute (MDWE) protections, aligning with the "PR_SET_MDWE"
and "PR_MDWE_REFUSE_EXEC_GAIN" functionality introduced in Linux kernel 6.3.
This feature establishes a stringent policy against creating memory mappings
that are _simultaneously writable and executable_, closely adhering to the
executable space protection mechanisms inspired by "PaX". In addition, Syd
fortifies these MDWE protections by employing kernel-level seccomp filters on
critical system calls, including _mmap_(2), _mmap2_(2), _mprotect_(2),
_pkey_mprotect_(2), and _shmat_(2). These filters are designed to intercept and
restrict operations that could potentially contravene MDWE policies, such as
attempts to make non-executable memory mappings executable or to map shared
memory segments with executable permissions. By integrating "PR_SET_MDWE" for
preemptive kernel enforcement and utilizing seccomp filters for granular,
kernel-level control over system call execution, Syd provides a robust defence
mechanism against exploitation techniques that exploit memory vulnerabilities,
thereby ensuring a securely hardened execution environment. This restriction may
be relaxed using the "trace/allow_unsafe_memory:true" sandbox command at startup.
Note even with this restriction relaxed, Syd is going to call "PR_SET_MDWE", but
it will use the "PR_MDWE_NO_INHERIT" flag to prevent propagation of the MDWE
protection to child processes on _fork_(2).

## Advanced Memory Protection Mechanisms

Syd version 3.15.1 enhances its security framework by integrating
sophisticated a seccomp BPF hook to meticulously block
_executable+shared_ memory mappings, targeting a critical vulnerability
exploitation pathway. As of version 3.21.3, Syd also blocks
_executable+anonymous_ memory. These updates refine the sandbox's
defence against unauthorised memory access and arbitrary code execution
by inspecting and filtering system calls, notably _mmap_(2), and
_mmap2_(2), to enforce stringent policies against dangerous memory
mapping combinations. While this bolstered security measure
significantly reduces the attack surface for exploits like buffer
overflows and code injections, it acknowledges potential legitimate use
cases, such as Just-In-Time (JIT) compilation and plugin architectures,
that may require exceptions. To accommodate necessary exceptions without
compromising overall security, Syd allows these restrictions to be
relaxed with explicit configuration through the
"trace/allow_unsafe_memory:true" command, ensuring that developers and
administrators can fine-tune the balance between security and
functionality according to specific requirements, with a keen eye on
preventing the propagation of relaxed security settings to child
processes.

## Null Address Mapping Prevention

In our ongoing effort to enhance the security features of Syd, as of version
3.15.1 we introduced a crucial update inspired by the practices of HardenedBSD,
specifically aimed at bolstering our sandbox's defences against null pointer
dereference vulnerabilities. Following the model set by HardenedBSD, Syd now
includes a new security measure that completely prohibits the mapping of memory
at the NULL address using the _mmap_(2) and _mmap2_(2) system calls with the
MAP_FIXED and MAP_FIXED_NOREPLACE flags. This addition is implemented through
meticulous seccomp filter rules that block these specific mapping requests when
the first argument (addr) is zero, effectively rendering attempts to exploit
null pointer dereferences as non-viable by ensuring such memory allocations
result in respective system call getting denied with EACCES. By disallowing the
execution of arbitrary code at the NULL address, Syd significantly reduces the
attack surface associated with such vulnerabilities, reinforcing the sandbox's
commitment to providing a robust security framework for Linux systems. This
technical enhancement reflects our dedication to leveraging advanced security
insights from the broader community, embodying our proactive stance on
safeguarding against evolving threats.

## Enhanced Security for Memory File Descriptors

In version 3.21.1, Syd significantly enhanced its security posture by
introducing restrictions on memory file descriptors (memfds). The
_memfd_create_(2) system call is now sandboxed under Read, Write, and
Exec controls, with the name argument prepended with "/memfd:" before
access checks. This allows administrators to globally deny access to
memfds using rules like "deny/read,write,exec+/memfd:\*". Additionally,
the _memfd_secret_(2) system call, which requires the
"secretmem.enable=1" boot option and is seldom used, was denied to
prevent potential exploits. Despite file I/O being restricted on secret
memfds, they could be abused by attackers to write payloads and map them
as executable, thus bypassing denylisted code execution controls.

Building on these changes, version 3.21.2 further fortifies security by
making memfds non-executable by default. This is achieved by removing
the "MFD_EXEC" flag and adding the "MFD_NOEXEC_SEAL" flag to
_memfd_create_(2), ensuring memfds cannot be made executable. Notably,
the "MFD_NOEXEC_SEAL" flag requires Linux-6.3 or newer to function.
These measures collectively mitigate the risk of memfd abuse, which can
involve executing malicious code within a sandbox, circumventing
security mechanisms like Exec, Force, and TPE sandboxing. For scenarios
where executable or secret memfds are genuinely required, the
"trace/allow_unsafe_memfd:1" option allows for relaxing these
restrictions, though it introduces increased security risks. By default,
these enhancements enforce a robust security posture, preventing
attackers from leveraging memfds as a vector for unauthorized code
execution.

## Path Masking

Introduced in version 3.16.7, the "Path Masking" feature in Syd enhances
security by enabling the obfuscation of file contents without denying
access to the file itself. This functionality is critical in scenarios
where compatibility requires file presence, but not file readability.
Path Masking works by redirecting any attempt to _open_(2) a specified
file to the character device "/dev/null", effectively presenting an
empty file to the sandboxed process. The original file metadata remains
unchanged, which is essential for applications that perform operations
based on this data. Moreover, masked files can still be executed,
providing a seamless integration where executability is required but
content confidentiality must be preserved.

This feature leverages _glob_(3p) patterns to specify which files to
mask, allowing for flexible configuration tailored to diverse security
needs. By default, Syd masks sensitive paths such as "/proc/cmdline" to
prevent the leakage of potentially sensitive boot parameters, aligning
with Syd's security-first design philosophy. Path Masking is a robust
security enhancement that minimises the risk of sensitive data exposure
while maintaining necessary system functionality and compliance with
expected application behaviors.

## Refined Socket System Call Enforcement

In Syd version 3.16.12, we have strengthened the enforcement of socket system
call restrictions within the sandbox using kernel-level BPF filters. This
enhancement builds upon existing features by embedding these controls directly
into the Syd process, ensuring that even if Syd is compromised, it cannot
utilise or manipulate denied socket domains. This proactive measure restricts
socket creation strictly to permitted domains such as UNIX (AF_UNIX), IPv4
(AF_INET), and IPv6 (AF_INET6), significantly reducing the network attack
surface. The "trace/allow_unsupp_socket:1" option allows for the extension of
permissible socket domains, catering to specific needs but potentially
increasing exposure risks. Additionally, "trace/allow_safe_kcapi:1" enables access
to the Kernel Crypto API, facilitating necessary cryptographic operations
directly at the kernel level. These enhancements provide a more secure and
configurable environment, allowing administrators precise control over network
interactions and improving the overall security posture of the sandbox.

## Enhanced Execution Control (EEC)

The Enhanced Execution Control (EEC) feature, introduced in Syd version
3.17.0, represents a significant advancement in the sandbox’s defence
mechanisms. This feature strategically disables the _execve_(2) and
_execveat_(2) system calls for the Syd process after they are no longer
required for executing the sandbox process, thus safeguarding against
their potential abuse by a compromised Syd process. The prohibition of
these critical system calls adds a robust layer to the existing
Memory-Deny-Write-Execute (MDWE) protections, intensifying the system’s
defences against exploit techniques such as code injection or
return-oriented programming (ROP). Concurrently, EEC ensures that the
_ptrace_(2) syscall is limited following the initial use of the
"PTRACE_SEIZE" call for execution-related mitigations. This action
effectively prevents subsequent system trace operations, barring
unauthorised process attachments and further securing the system against
manipulation. Together, these measures enhance Syd's security
architecture, reflecting an ongoing commitment to implement rigorous,
state-of-the-art safeguards within the execution environment.

As of version 3.17.1, the Enhanced Execution Control (EEC) has been
further strengthened by integrating _mprotect_(2) hardening mechanisms
specifically targeting the prevention of the "ret2mprotect" exploitation
technique. This enhancement blocks attempts to alter memory protections
to executable (using the "PROT_EXEC" flag) via the _mprotect_(2) and
_pkey_mprotect_(2) system calls. By adding these checks, EEC mitigates the
risk associated with compromised Syd processes by enforcing stringent
memory operation policies that prevent unauthorised memory from becoming
executable, thereby countering sophisticated memory corruption attacks
such as return-oriented programming (ROP) and other code injection
strategies. This proactive security measure is crucial for maintaining
the integrity of the sandbox environment, ensuring that Syd continues to
offer robust protection against evolving exploit techniques.

## Enhanced Path Integrity Measures

As of version 3.17.4, Syd incorporates crucial enhancements to maintain
the integrity of file system paths by systematically denying and masking
paths that contain control characters. These modifications are essential
for preventing the exploitation of terminal-based vulnerabilities and
for maintaining robustness in logging activities. Paths identified with
control characters are not only denied during sandbox access check but
are also sanitized when logged to ensure that potentially harmful data
does not compromise log integrity or facilitate inadvertent security
breaches. Such measures underscore Syd's ongoing commitment to
fortifying security by adhering to rigorous, up-to-date standards for
handling untrusted input efficiently.

As of version 3.18.6, this restriction can be relaxed by using the
setting "trace/allow_unsafe_filename:1"

## Device Sidechannel Mitigations

As of Syd version 3.21.0, Syd's device sidechannel mitigations align
closely with "GRKERNSEC_DEVICE_SIDECHANNEL" in Grsecurity, aiming to
prevent timing analyses on block or character devices via _stat_(2) or
_inotify_(7)/_fanotify_(7). For stat-related system calls, Syd, like
Grsecurity, matches the last access and modify times to the creation
time for devices, thwarting unprivileged user timing attacks. Instead of
dropping events, Syd strips access and modify _fanotify_(7)/_inotify_(7)
flags at syscall entry, preventing unsafe _fanotify_(7)/_inotify_(7)
event generation. This approach ensures unauthorized users cannot
determine sensitive information, such as the length of the administrator
password. Syd's solution offers robust security by dynamically stripping
flags, enhancing protection against these sidechannel attacks without
compromising functionality.

Refer to the following links for more information:

- https://web.archive.org/web/20130111093624/http://vladz.devzero.fr/013_ptmx-timing.php
- https://en.wikibooks.org/wiki/Grsecurity/Appendix/Grsecurity_and_PaX_Configuration_Options#Eliminate_stat/notify-based_device_sidechannels

## Restricting CPU Emulation System Calls

As of version 3.22.1, Syd denies the _modify_ldt_(2), _subpage_prot_(2),
_switch_endian_(2), _vm86_(2), and _vm86old_(2) system calls by default,
which are associated with CPU emulation functionalities. These calls can
only be allowed if the "trace/allow_unsafe_cpu" option is explicitly
set. This restriction helps mitigate potential vulnerabilities and
unauthorized access that can arise from modifying CPU state or memory
protections, thus strengthening the overall security posture of the
sandbox environment.

## Kernel Keyring Access Restriction

To enhance system security, access to the kernel's key management
facility via the _add_key_(2), _keyctl_(2), and _request_key_(2) system
calls is restricted by default as of version 3.22.1. These calls are
crucial for managing keys within the kernel, enabling operations such as
adding keys, manipulating keyrings, and requesting keys. The restriction
aims to prevent unauthorized or potentially harmful modifications to
keyrings, ensuring that only safe, controlled access is permitted.
However, administrators can relax this restriction by enabling the
"trace/allow_unsafe_keyring" option, allowing these system calls to be
executed when necessary for legitimate purposes.

## Restricting Memory Protection Keys System Calls

As of version 3.22.1, Syd denies the system calls _pkey_alloc_(2),
_pkey_free_(2), and _pkey_mprotect_(2) by default. These system calls
are associated with managing memory protection keys, a feature that can
be leveraged to control memory access permissions dynamically. To allow
these system calls, administrators can enable the
"trace/allow_unsafe_pkey" option. This restriction enhances security by
preventing unauthorized or potentially harmful manipulations of memory
access permissions within the sandbox environment, ensuring stricter
control over memory protection mechanisms.

# Force Close-on-Exec

The "trace/force_cloexec" option, introduced in Syd version 0.22.1,
ensures that all _open_(2), _openat_(2), and _openat2_(2) system calls
made by the sandbox process include the "O_CLOEXEC" flag. This feature
can be toggled at runtime via Syd's virtual stat API, enabling dynamic
adjustment of confinement levels as needed. The "O_CLOEXEC" flag, when
set on file descriptors, ensures they are automatically closed when
executing a new program via _execve_(2) or similar system calls. This
automatic closure of file descriptors is critical for enhancing security
and safety, as it prevents file descriptors from being unintentionally
inherited by newly executed programs, which could otherwise lead to
unauthorized access to sensitive files or resources. By enforcing the
"O_CLOEXEC" flag across all _open_(2) calls, Syd mitigates the risk of
file descriptor leakage, effectively isolating the sandboxed environment
and ensuring a clean execution context for newly spawned processes. This
feature thus fortifies the security boundary of the sandbox, ensuring
that potentially sensitive file descriptors are not inadvertently
exposed to less privileged or potentially malicious code executed within
the sandbox.

# HISTORY & DESIGN

- *sydbox-0* https://git.sr.ht/~alip/syd/tree/sydbox-0 is a _ptrace_(2) based sandbox.
- *sydbox-1* https://git.sr.ht/~alip/syd/tree/sydbox-1 is a _ptrace_(2) and _seccomp_(2) based sandbox.
- *sydbox-2* https://git.sr.ht/~alip/syd/tree/sydbox-1 is a _seccomp_(2) and _seccomp-notify_ based sandbox.
- *sydbox-3* is a rewrite of *sydbox-2* in Rust and it's what you are looking at.

This codebase has a history of a bit over 15 years and up to this point we have
used C11 as our implementation language for various reasons. With *sydbox-3*
we are moving forwards one step and writing the sandbox from scratch using the
Rust programming language with the only non-Rust dependency being libseccomp.
Although we inherit many ideas and design decisions from the old codebase, we
also don't shy away from radically changing the internal implementation making
it much simpler, idiomatic, and less prone to bugs. We have _proper multiarch
support_ since release 3.0.11, e.g on x86-64, you can run your x32 or x86
binaries just fine under Syd.

This version takes advantage of multithreading and handles system calls
using a thread pool whose size is equal to the number of CPUs on the
running machine and utilises globsets to match a list of patterns at
once, thus continues to perform reasonably well even with very long
rulesets. Moreover, access violations are logged into _syslog_(3), so
you may use a command like "journalctl SYSLOG_IDENTIFIER=syd" (or
shortly _syd-log_(1)) to view the sandbox logs. This version also comes
with four new sandboxing types called *Lock Sandboxing*, *Memory
Sandboxing*, *PID sandboxing*, *Stat Sandboxing*, *Force Sandboxing*:
*Lock Sandboxing* utilises the Landlock Linux Security Module (LSM),
*Memory Sandboxing* allows the user to define a per-process memory
limit, *PID sandboxing* allows the user to define a limit on the maximum
number of running tasks under the sandbox, *Stat Sandboxing* can be used
to effectively _hide files and directories_ from the sandboxed process
whereas *Force Sandboxing* can be used to verify file checksums prior to
exec, similar to HardenedBSD's Integriforce and NetBSD's Veriexec.

Finally, the new Syd has support for namespaces. Use e.g. "syd -m
unshare/user:1" to create a user namespace. You may use "mount", "uts", "ipc",
"pid", "net", and "cgroup" instead of "user" to create various namespaces. You
may use the "container" profile as a shorthand to create namespaces with
"syd -pcontainer".

You may use Syd as your login shell because it is very practical to have a
restricted user. To do this simply add "/path/to/syd" to the file "/etc/shells"
and do "chsh -s /path/to/syd username" as root. In this mode the sandbox may be
configured using the files "/etc/user.syd-3" and "~/.user.syd-3". If you want to
restrict user configuration of the sandbox, lock the sandbox using "lock:on" at
the end of the site-wide configuration file.

# EXHERBO

Syd is the default sandbox of *Exherbo Linux*. We use it to provide a restricted
environment under which package builds run with controlled access to file system
and network resources. *exheres-0* has a function called *esandbox* to interact
with Syd.

# BUGS

## TOCTOU due to pointer modification

In the operation of Syd, certain system calls are not fully emulated due
to _seccomp_(2) limitations, resulting in the sandbox process continuing
these calls directly. These include _execve_(2), _execveat_(2) for
execution, _chdir_(2), _fchdir_(2) for directory changes, and _open_(2)
operations with *O_PATH* flag. Consequently, this behavior exposes
vulnerabilities to *time-of-check to time-of-use* attacks, allowing for
the circumvention of *Exec Sandboxing* and *Force Sandboxing* to execute
denylisted paths, the bypass of *Stat Sandboxing* for unauthorised
directory access _without disclosing directory contents_ (owing to
_getdents_(2) call emulation), and the detection of hidden files
_without revealing file metadata_, as _stat_(2) calls are emulated. See
https://bugzilla.kernel.org/show_bug.cgi?id=218501 for more information.

The user is recommended to apply the "write xor execute" (W^X) approach
when writing sandbox rules such that no part of the filesystem is both
writable and executable at the same time. *Write Sandboxing*, *Lock
Sandboxing*, or the *noexec* option with a bind mount can be used for
this as all of these options are TOCTOU-free.

As of version 3.16.0, Syd applies an effective TOCTOU mitigation for
_execve_(2), and _execveat_(2) system calls utilizing
"PTRACE_EVENT_EXEC". To mitigate, Syd stops the process before return
from _execve_(2) using _ptrace_(2) and checks the updated "/proc/pid/exe"
magic symbolic link of the process against the sandbox. If the path is
found to be denylisted, which indicates a successful TOCTOU attempt to
execute a denylisted path, Syd kills the process.

As of version 3.17.0, Syd denylists _execve_(2), and _execveat_(2) system calls
in the Syd process with a seccomp-bpf filter which is configured to terminate
Syd if these system calls are ever attempted, after the initial execution of the
sandbox process. Read "Enhanced Execution Control (EEC)" section for more
information.

As of version 3.19.0, Syd turns the "O_PATH" flag in _open_(2) system
call arguments to the "O_RDONLY" flag and emulates the system call as
usual which avoids the TOCTOU vector. This may have unintended side
effects (e.g. _open_(2) with "O_PATH" will not update access time but
_open_(2) with "O_RDONLY" will) so the option
"trace/allow_unsafe_open:1" is provided to avoid the flag change and
continue the _open_(2) call in the sandbox process.

## TOCTOU due to filesystem modification

Networking system calls, _bind_(2), _connect_(2), _sendto_(2),
_sendmsg_(2), and _sendmmsg_(2) are vulnerable to TOCTOU due to
filesystem modification. Time of Check to Time of Use (TOCTOU)
vulnerabilities occur when the state of a system or resource changes
between the verification stage and the use stage, leading to potential
security issues. In the context of filesystem operations, this
vulnerability happens when the attributes or state of a file (like a
UNIX socket) change unexpectedly after being checked by system calls
such as _bind_(2) or _connect_(2) but before these calls are executed.
This discrepancy can lead to unauthorized access or incorrect system
behavior. Specifically, there is a noted challenge with UNIX sockets
where there is currently no method available to open a socket in a
manner that is guaranteed to be secure (atomic) against modifications
before it is bound or connected. This limitation necessitates a cautious
approach and future mitigation planning.

## Uninterruptible FIFOs

Due to the way seccomp-notify and signal handling interacts, FIFOs are
not interruptible under Syd. This means e.g. if an _open_(2) system call
is blocked for reading on a FIFO, sending the sandbox process that made
the original _open_(2) system call an interrupt signal will be queued by
the Linux kernel and _not_ delivered to the sandbox process until the
system call is handled by Syd. Therefore only opening the write end of
the FIFO can unblock the original blocked _open_(2) system call. This is
safe and secure, however it also breaks expectations of some programs
which rely on FIFOs being interruptible.

# SEE ALSO

_syd_(1), _syd_(2), _syd_(5), _seccomp_(2), _pidfd_getfd_(2), _pidfd_send
signal_(2), _ioctl_(2), _ioctl_tty_(2), _prctl_(2), _namespaces_(7),
_cgroup_namespaces_(7), _ipc_namespaces_(7), _mount_namespaces_(7),
_network_namespaces_(7), _pid_namespaces_(7), _user_namespaces_(7),
_uts_namespaces_(7)

- https://exherbolinux.org/docs/eapi/exheres-for-smarties.html#sandboxing

# AUTHORS

Maintained by Ali Polatel. Up-to-date sources can be found at
https://gitlab.exherbo.org/sydbox/sydbox.git and bugs/patches can be
submitted to https://gitlab.exherbo.org/groups/sydbox/-/issues. Discuss
in #sydbox on Libera Chat.
