//
// Syd: rock-solid unikernel
// src/syd-key.rs: Utility to generate AES-CTR Key and IV
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{env, process::ExitCode};

use anyhow::{Context, Result};
use const_hex::encode;
use getargs::{Opt, Options};
use syd::hash::Key;

fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));

    #[allow(clippy::disallowed_methods)]
    #[allow(clippy::never_loop)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    let key = Key::random()
        .map_err(anyhow::Error::msg)
        .context("Error generating key")?;
    println!("crypt/key:{}", encode(&key));

    Ok(ExitCode::SUCCESS)
}

fn help() {
    println!("Usage: syd-key [-h]");
    println!("Utility to generate 256-bit AES-CTR key");
    println!("  -h        Print this help message and exit");
}
